package mods.immibis.infinitubes;

import net.minecraft.block.Block;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.inventory.IInventory;
import net.minecraft.inventory.ISidedInventory;
import net.minecraft.inventory.InventoryLargeChest;
import net.minecraft.item.ItemStack;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.tileentity.TileEntityChest;
import net.minecraft.world.World;
import net.minecraftforge.common.util.ForgeDirection;

public class InventoryHelper {
	public static IInventory getInventory(World world, int x, int y, int z, int side) {
		TileEntity te = world.getTileEntity(x, y, z);
		if(te instanceof ISidedInventory) {
			ISidedInventory isi = (ISidedInventory)te;
			int[] slots = isi.getAccessibleSlotsFromSide(side);
			if(slots.length == 0)
				return null;
			return new VanillaSidedAdapter(isi, slots, side);
		}
		if(te instanceof TileEntityChest) {
			IInventory inv = (IInventory)te;
			Block block = world.getBlock(x, y, z);
			for(int k = 2; k < 6; k++) {
				ForgeDirection fd = ForgeDirection.VALID_DIRECTIONS[k];
				if(!world.blockExists(x+fd.offsetX, y, z+fd.offsetZ))
					continue;
				if(world.getBlock(x+fd.offsetX, y, z+fd.offsetZ) != block)
					continue;
				te = world.getTileEntity(x + fd.offsetX, y, z + fd.offsetZ);
				if(te instanceof TileEntityChest) {
					inv = new InventoryLargeChest("", inv, (IInventory)te);
				}
			}
			return inv;
		}
		if(te instanceof IInventory)
			return (IInventory)te;
		return null;
	}
	
	private static class VanillaSidedAdapter implements IInventory, IRestrictedInventory {
		private ISidedInventory inv;
		private int side;
		private int[] slots;
		
		public VanillaSidedAdapter(ISidedInventory inv, int[] slots, int side) {
			this.inv = inv;
			this.slots = slots;
			this.side = side;
		}

		@Override
		public int getSizeInventory() {
			return slots.length;
		}

		@Override
		public ItemStack getStackInSlot(int i) {
			return inv.getStackInSlot(slots[i]);
		}

		@Override
		public ItemStack decrStackSize(int i, int j) {
			return inv.decrStackSize(slots[i], j);
		}

		@Override
		public ItemStack getStackInSlotOnClosing(int i) {
			return inv.getStackInSlotOnClosing(slots[i]);
		}

		@Override
		public void setInventorySlotContents(int i, ItemStack itemstack) {
			inv.setInventorySlotContents(slots[i], itemstack);
		}

		@Override
		public String getInventoryName() {
			return inv.getInventoryName();
		}

		@Override
		public boolean hasCustomInventoryName() {
			return inv.hasCustomInventoryName();
		}

		@Override
		public int getInventoryStackLimit() {
			return inv.getInventoryStackLimit();
		}

		@Override
		public void markDirty() {
			inv.markDirty();
		}

		@Override
		public boolean isUseableByPlayer(EntityPlayer entityplayer) {
			return inv.isUseableByPlayer(entityplayer);
		}

		@Override
		public void openInventory() {
			inv.openInventory();
		}

		@Override
		public void closeInventory() {
			inv.closeInventory();
		}

		@Override
		public boolean isItemValidForSlot(int i, ItemStack itemstack) {
			return inv.isItemValidForSlot(slots[i], itemstack);
		}

		@Override
		public boolean canExtract(int slot) {
			slot = slots[slot];
			
			ItemStack stack = inv.getStackInSlot(slot);
			return inv.canExtractItem(slot, stack, side);
		}

		@Override
		public boolean canInsert(int slot, ItemStack item) {
			return inv.isItemValidForSlot(slots[slot], item) && inv.canInsertItem(slots[slot], item, side);
		}
	}
		

	public static IInventory getInventoryOnSide(World worldObj, int xCoord, int yCoord, int zCoord, int direction) {
		ForgeDirection fd = ForgeDirection.VALID_DIRECTIONS[direction];
		xCoord += fd.offsetX;
		yCoord += fd.offsetY;
		zCoord += fd.offsetZ;
		return getInventory(worldObj, xCoord, yCoord, zCoord, direction ^ 1);
	}
}
